#    PyNomo - nomographs with Python
#    Copyright (C) 2007-2015  Leif Roschier
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" PyNomo - nomographs or nomograms with Python

Pynomo is a program to create (pdf) nomographs (nomograms)
using Python interpreter. A nomograph (nomogram) is a graphical
solution to an equation.
"""

__all__ = ["nomo_axis", "nomo_axis_func",
           "nomo_grid_box", "nomo_grid", "nomo_wrapper", "nomographer",
           "isopleth", "nomograph3"]
